/*
	PROforma, PROGS FOnt & Raster MAnager

	Character display routine for PROforma.

	PFShowChar gstate, unicode

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 19 August 1993
	17.08.1994 adjusted to use the CharPath iterator
	21.09.1995 proper use of MUTEX library for accessing fontcache
	01.10.1995 adjusted size of bbox, now expanded by half a pixel in each
			   direction. This gives better results
	03.10.1995 clipping now done in DrawMask which may or may not support it
	12.10.1995 use new, faster cache (font specific instead of global)
	14.10.1995 charwidth now stored on fontcache scaled to fontsize (faster)
	18.10.1995 charwidth is now returned at real size by PFCharPathEl
	16.12.1995 now always assume that DrawMask can clip !!
	04.11.1996 font quality improved by using fixed flatness of 1/32 when
			   rendering characters
	25.11.1996 added a second pass to render char to make sure small vertical
			   spans are also drawn.
	27.11.1996 added code to allow PFPWidth commands to get minimal pixel
			   widths for characters, by introducing noDisplayChars
*/

#include "gstate.h"
#include "system.h"
#include "cache.h"
#include "path.h"
#include "hint.h"
#include "list.h"			/* for definition of min and max */

/* for drawing in cache */
Error CacheHorizontal(Bitmap mask, short y, short xs, short xe);
Error CacheVertical(Bitmap mask, short x, short y, short ye);

/* move an entire path, both coordinates are in pixels (short, NOT fixpoint) */
static void fixpath(Gstate gstate, int x, short y)
{
	segment *el=gstate->path->next;
	x=short2pt(x);
	while (el->next)
	{
		if (!el->isclip)
		{
			el->startx+=x;
			el->starty+=y; el->endy+=y;
		}
		el=el->next;
	}
}

static Error NCDrawChar(Gstate gstate)
{
	if (!gstate->noDisplayChars)
	{
		/* to make the coordinates correct, add posdev to each pair */
		fixpath(gstate,pt2rshort(gstate->posdev.x),pt2rshort(gstate->posdev.y));

		/* and draw it */
		fillpath(gstate,gstate->raster->bitdraw->TrapezoidHorizontal,
			gstate->raster,PF_WINDING_RULE);
	}
	gstate->chwx=short2pt(gstate->boxxmax+1);
	return ERR_OK;
}

static Error DrawChar(Gstate gstate)
{
	if (!gstate->noDisplayChars)
	{
		/* to make the coordinates correct, add posdev to each pair */
		fixpath(gstate,pt2rshort(gstate->posdev.x),pt2rshort(gstate->posdev.y));

		/* and draw it */
		cfillpath(gstate,gstate->raster->bitdraw->LineHorizontal,gstate->raster);
	}
	gstate->chwx=short2pt(gstate->boxxmax+1);
	return ERR_OK;
}

static Error DrawCachedChar(Gstate gstate, int code)
{
	CachedMask *c;
	Error err;

	if (ERR_OK==(err=CacheFind(gstate->font, code, &c)))
	{
		/* draw character */
		if (!gstate->noDisplayChars)
		{
			err=gstate->raster->bitdraw->DrawMask(gstate->raster, &c->mask,
									  pt2rshort(gstate->posdev.x)-c->xoff,
									  pt2rshort(gstate->posdev.y)-c->yoff);
			if (err) return err;
		}
		gstate->boxxmin=-c->xoff;
		gstate->boxymin=-c->yoff;
		gstate->boxxmax=c->mask.xsiz-c->xoff;
		gstate->boxymax=c->mask.ysiz-c->yoff;
		gstate->CharWidth.x=c->CharWidth.x;
		gstate->CharWidth.y=c->CharWidth.y;
		gstate->chwx=c->chwx;
	}
	return err;
}

static Error DrawNewCachedChar(Gstate gstate, int code, int xsiz, int ysiz)
{
	CachedMask *c, *cc;
	coor posdev;
	segment *skeep;

	/* have to be able to create a new cache entry */
	if (CacheCreate(gstate->font, code, xsiz, ysiz, &cc))
		return ERR_NC;
	c=cc;

	/* fill in new cache entry */

	c->xoff=-gstate->boxxmin;
	c->yoff=-gstate->boxymin;
	c->CharWidth.x=gstate->CharWidth.x;
	c->CharWidth.y=gstate->CharWidth.y;
	gstate->chwx=c->chwx=short2pt(gstate->boxxmax+1);

	fixpath(gstate,c->xoff,c->yoff);
	cfillpath(gstate,CacheHorizontal,(Bitmap)&c->mask);
	skeep=gstate->path;
	gstate->path=gstate->path2;
	gstate->path2=skeep;
	fixpath(gstate,c->yoff,c->xoff);
	cfillpath(gstate,CacheVertical,(Bitmap)&c->mask);

	/* draw character */
	if (!gstate->noDisplayChars)
		gstate->raster->bitdraw->DrawMask(gstate->raster, &c->mask,
							  pt2rshort(gstate->posdev.x)-c->xoff,
							  pt2rshort(gstate->posdev.y)-c->yoff);
	return ERR_OK;
}

Error PFShowChar(Gstate gstate, int code)
{
	Error err=ERR_ITNF;

	if (gstate->usecache)
	{
		if (MUTEXCall(gstate->CacheForce,DrawCachedChar,2,gstate,code))
		{
			/* character was not yet in cache (ERR_ITNF) */

			int xsiz,ysiz;

			PathNoClip(gstate);
			CalcChar(gstate, code);

			xsiz=gstate->boxxmax-gstate->boxxmin;
			ysiz=gstate->boxymax-gstate->boxymin;

			if (MUTEXCall(gstate->CacheForce,DrawNewCachedChar,
							4, gstate,code,xsiz,ysiz))
				return DrawChar(gstate);
		}
	}
	else
	{
		PFPathClear(gstate);
		CalcChar(gstate, code);
		err=NCDrawChar(gstate);
	}
	return err;
}
