/*
	PROforma, PROGS FOnt & Raster MAnager

	Show and Width operators for strings.
	The backslash ('\') is interpretted as escape character. Between to
	backslashes you can give the name or code (decimal) of the character to
	print. A backslash can be included with the '\\' sequence.

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 14 April 1995
	15.10.1995 use specific PFMoveR -> faster, init gstate->font here
	25.11.1995 current point no longer advanced by show operators -> fixed
	27.11.1996 PFPStringShow & PFPStringWidth added
	03.08.1997 - PFStringShow(kern)Just fixed, used to loose characters
			   - PFStringShow(Kern)Just now properly uses ExtraEos when
				 counting spaces
	27.11.1997 - PFStringShowTrack now uses ExtraEOS and uicode replacement
				 when counting characters
			   - problem with counting chars using StringContinues macro fixed
*/

#include "gstate.h"
#include "system.h"
#include "cache.h"

#define MoveCursor	reluser2dev(gstate,w); \
					gstate->posdev.x+=w->x; \
					gstate->posdev.y+=w->y
#define Done		gstate->subpathorg=gstate->posdev; \
					return PFPathClear(gstate)

Error PFStringShow(Gstate gstate, const char *string)
{
	coor *w=&gstate->CharWidth;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		MoveCursor;
	}
	Done;
}

Error PFStringShowKern(Gstate gstate, const char *string)
{
	coor *w=&gstate->CharWidth;
	short code, next;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		PFShowChar(gstate, code);
		next=PFChar2UniCode(gstate,string,&string);
		KernAdjustWidth(gstate, code, next);
		MoveCursor;
		code=next;
	}
	Done;
}

Error PFStringShowX(Gstate gstate, const char *string, const pt *Xlist)
{
	coor *w=&gstate->CharWidth;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		w->x=*Xlist++;
		w->y=0;
		MoveCursor;
	}
	Done;
}

Error PFStringShowXY(Gstate gstate, const char *string, const pt *Xlist, const pt *Ylist)
{
	coor *w=&gstate->CharWidth;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		w->x=*Xlist++;
		w->y=*Ylist++;
		MoveCursor;
	}
	Done;
}

Error PFStringShowTrack(Gstate gstate, const char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	const char *s=string;
	pt add=0, wdt; short len=0, code;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	/* determine length of string */
	for (s=string,len=0 ;
		 code=PFChar2UniCode(gstate,s,&s),
		 StringContinues(gstate,code) ; )
		len++;
	PFStringWidth(gstate,string,&wdt);
	if (len) add=(width-wdt)/(len-1);

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		w->x+=add;
		MoveCursor;
	}
	Done;
}

Error PFStringShowJust(Gstate gstate, const char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	const char *s;
	pt add=0, wdt;
	short code;
	int len=0;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	/* determine length of string */
	for (s=string,len=0 ;
		 code=PFChar2UniCode(gstate,s,&s),
		 StringContinues(gstate,code) ; )
		if (code==' ') len++;
	PFStringWidth(gstate,string,&wdt);
	if (len) add=(width-wdt)/len;

	while (StringContinues(gstate,*string))
	{
		code=*string;
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		if (code==' ') w->x+=add;
		MoveCursor;
	}
	Done;
}

Error PFStringShowKernJust(Gstate gstate, const char *string, pt width)
{
	coor *w=&gstate->CharWidth;
	const char *s;
	pt add=0, wdt;
	short code, next;
	int len=0;

	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	/* determine length of string */
	for (s=string,len=0 ;
		 code=PFChar2UniCode(gstate,s,&s),
		 StringContinues(gstate,code) ; )
		if (code==' ') len++;
	PFStringWidthKern(gstate,string,&wdt);
	if (len) add=(width-wdt)/len;

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		PFShowChar(gstate, code);
		next=PFChar2UniCode(gstate,string,&string);
		KernAdjustWidth(gstate, code, next);
		if (code==' ') w->x+=add;
		MoveCursor;
		code=next;
	}
	Done;
}

Error PFStringWidth(Gstate gstate, const char *string, pt *xwidth)
{
	int width=0;

	*xwidth=0;
	if (!string) return ERR_OK;

	while (StringContinues(gstate,*string))
	{
		width+=CharWidth(gstate, PFChar2UniCode(gstate,string,&string));
	}
	*xwidth=width*(gstate->fontsize/1000);
	return ERR_OK;
}

Error PFStringWidthKern(Gstate gstate, const char *string, pt *xwidth)
{
	int width=0;
	short code, next;

	*xwidth=0;
	if (!string) return ERR_OK;

	code=PFChar2UniCode(gstate,string,&string);
	while (StringContinues(gstate,code))
	{
		width+=CharWidth(gstate, code);
		width+=KernAdjust(gstate, *string, *(string+1));
		next=PFChar2UniCode(gstate,string,&string);
		width+=KernAdjust(gstate, code, next);
		code=next;
	}
	*xwidth=width*(gstate->fontsize/1000);
	return ERR_OK;
}

Error PFPStringShow(Gstate gstate, const char *string)
{
	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	if (!gstate->font->mwidth)
	{
		gstate->noDisplayChars=TRUE;
		PFShowChar(gstate, 'm');
		gstate->noDisplayChars=FALSE;
	}

	while (StringContinues(gstate,*string))
	{
		PFShowChar(gstate, PFChar2UniCode(gstate,string,&string));
		gstate->posdev.x+=gstate->chwx;
	}
	Done;
}

Error PFPStringWidth(Gstate gstate, const char *string, short *xwidth)
{
	short width=0;
	*xwidth=0;
	if (!string) return ERR_OK;
	if (!gstate->font) CacheFontFind(gstate);

	while (StringContinues(gstate,*string))
		width+=PCharWidth(gstate, PFChar2UniCode(gstate,string,&string));
	*xwidth=width;
	return ERR_OK;
}
