/*
	g3 fax file encoder for Qfax,
	code originally received from Jonathan Hudson
	modified on 15.03.1997 by Joachim Van der Auwera for
	- reentrant access
	- use syslib and don't rely on standard c libraries (for xmod)
*/

#include <io.h>
#include <time.h>

#include "driv/faxqheader.h"
#include "driv/faxg3encoder.h"
#include "driv/faxalphabet.h"

/* Get next changing element (2-D) */

static int get_nce(CODERSTATE *code, int bp)
{
	int cu_co;

	cu_co = ((code->refline[bp >> 3] & code->mbits[bp & 0x07]) != 0);

	code->lce_col = !cu_co;

	if (cu_co)
	{
	while (((code->refline[bp >> 3] & code->mbits[bp & 0x07]))
					 && (bp < code->pix))
	{
		bp++;
	}
	}
	else
	{
	while ((!(code->refline[bp >> 3] & code->mbits[bp & 0x07]))
					 && (bp < code->pix))
	{
		bp++;
	}
	}
	return bp;
}


static void enter_bits (CODERSTATE *code, char *str)
{

	while (*str)
	{
	if (*str == '1')
	{
		code->outbuf[code->bitnr >> 3] |= code->mbits[code->bitnr & 0x07L];
	}
	code->bitnr++;
	str++;
	}
}

int FaxFlushBuf(CODERSTATE *code, short ws)
{
	int nw;

	IOPutBytes(code->chid, ws, (char *)code->outbuf, &nw);
	if (nw>=0) nw=ws;
	MEMClear(OUTBUFSIZE,code->outbuf);
	code->bitnr = 0;
	return nw;
}

static int CoderEOL(CODERSTATE *code)
{
	int ws, nw;

	enter_bits (code,EOL);
	if (code->pack)
	{
	if (code->line_2d)
	{
		if (!(--code->cnt_2d))
		{
		code->line_2d = 0;
		}
	}
	else
	{
			code->cnt_2d = code->cnt_ref;
		code->line_2d = 1;
	}
	if (code->line_2d)
	{
		enter_bits (code,"0");
	}
	else
	{
		enter_bits (code,"1");
	}
	}
	code->scan++;
	ws = (code->bitnr >> 3);
	if (code->bitnr & 0x07L)
	{
	if (ws > (OUTBUFSIZE / 2))
	{
		IOPutBytes(code->chid, ws, (char *)code->outbuf, &nw);
		if (nw>=0) nw=ws;
		code->outbuf[0] = code->outbuf[ws + 1];
		MEMClear(OUTBUFSIZE-1, (char *) code->outbuf + 1);
		code->bitnr &= 0x07L;
	}
	else
	{
		nw = ws;
	}
	}
	else
	{
		nw = FaxFlushBuf(code,ws);
	}
	return (nw != ws);
}

void FaxCloseFile(CODERSTATE *code)
{
	FaxFlushBuf(code,((code->bitnr+7) >> 3));
	IOClose(code->chid);
	code->chid=CHANNEL_NONE;
}

static int enter_white(CODERSTATE *code,int anz)
{
	if (anz > 63)
	{
	enter_bits (code,mu_code_white[(anz >> 6) - 1]);
	anz &= 0x3F;
	}
	enter_bits (code,white_codes[anz]);

	return 1;
}

static int enter_black(CODERSTATE *code,int anz)
{

	if (anz > 63)
	{
	enter_bits (code,mu_code_black[(anz >> 6) - 1]);
	anz &= 0x3F;
	}
	enter_bits (code,black_codes[anz]);

	return 0;
}

static int coder_2d(CODERSTATE *code, uchar *pixptr)
{
	int i, a1, dist, cc_cnt, j, u;
	int cur_col = 0, old_cc;

	i = 0;
	while (i < code->pix)
	{
	code->b1 = get_nce (code,i); /* next changing element on refline */
	if (code->lce_col == cur_col)
	{
		code->b1 = get_nce (code,code->b1);
	}

	a1 = i;

	if (cur_col)
	{
		while (((pixptr[a1 >> 3] & code->mbits[a1 & 0x07])) &&
											 (a1 < code->pix))
		{
		a1++;
		}
	}
	else
	{
		while ((!(pixptr[a1 >> 3] & code->mbits[a1 & 0x07])) &&
											(a1 < code->pix))
		{
		a1++;
		}
	}

	if (a1 == code->b1)
	{
		enter_bits (code,VERTICAL);
		cur_col = !cur_col;
		i = a1;
	}
	else
	{
		code->b2 = get_nce (code,code->b1);
		if (code->b2 < a1)
		{           /* pass mode */
		i = code->b2;
		enter_bits (code,PASSMODE);
		}
		else if ((dist = (abs (code->b1 - a1))) <= 3)
		{
		if ((code->b1 - a1) < 0)
		{
			enter_bits (code,vertical_r[dist - 1]);
		}
		else
		{
			enter_bits (code,vertical_l[dist - 1]);
		}
		cur_col = !cur_col;
		i = a1;
		}
		else
		{           /* horizontal mode */

		enter_bits (code,HORMODE);
		cc_cnt = 0;

		for (u = 0; u < 2; u++)
		{
			old_cc = cur_col;
			for (j = i; (j < code->pix); j++)
			{
			if ((pixptr[j >> 3]) & (code->mbits[j & 0x07]))
			{
				if (cur_col)
				cc_cnt++;
				else
				{
				cur_col = enter_white (code,cc_cnt);
				cc_cnt = 0;
				}
			}
			else
			{
				if (cur_col)
				{
				cur_col = enter_black (code,cc_cnt);
				cc_cnt = 0;
				}
				else
				cc_cnt++;
			}
			if (old_cc != cur_col)
			{
				break;
			}
			}
			i = j;
		}		/* end coding a0->a1, a1->a2 */
		if (cc_cnt)
		{
			if (cur_col)
			cur_col = enter_black (code,cc_cnt);
			else
			cur_col = enter_white (code,cc_cnt);
		}
		}
	}
	}
	return 0;
}
static void ClearCoder(CODERSTATE *code)
{
	if (code)
	{
		MEMClear(OUTBUFSIZE, code->outbuf);
		code->outptr = code->outbuf;
		code->scan = code->bytenr = 0;
		code->bitnr = 0;
		code->line_2d = 0;
		code->cnt_ref = (code->res) ? K_CONST_H : K_CONST_L;
		code->cnt_2d = code->cnt_ref;
	}
}
void FaxBeginPage(CODERSTATE *code)
{
	if (!code) return;
	ClearCoder(code);
	enter_bits(code,EOL);
	if (code->pack)
	{
		enter_bits(code,"1");
		code->line_2d = 0;
	}
}

int FaxCodeLine(CODERSTATE *code, uchar *pixptr)
{
	int i, curr_col = 0, cc_count = 0;

	if (code->pack && code->line_2d)
	{
		coder_2d(code,pixptr);
	}
	else
	{
		for (i = 0; i < code->pix; i++)
		{
			if ((pixptr[i >> 3]) & (code->mbits[i & 0x07]))
			{
				if (curr_col)
					cc_count++;
				else
				{
					curr_col = enter_white(code,cc_count);
					cc_count = 1;
				}
			}
			else
			{
				if (curr_col)
				{
					curr_col = enter_black(code,cc_count);
					cc_count = 1;
				}
				else
					cc_count++;
			}
		}

		if (cc_count)
		{
			if (curr_col)
				enter_black (code,cc_count);
			else
				enter_white (code,cc_count);
		}
	}

	while (code->bitnr & 0x07)
	{
	code->bitnr++;
	}

	code->bitnr += (12 - code->pack);

	if (code->pack)
	{
		MEMCopy(((code->pix) >> 3),pixptr,code->refline);
	}
	return CoderEOL (code);
}

void FaxCodeEOP(CODERSTATE *code)
{
	short i;

	/*
	 * Only do FIVE for RTC as we've just done one
	 */

	if(code->scan > code->maxlines)
	{
		code->maxlines = code->scan;
	}
	for (i = 0; i < 5; i++)
	{
		CoderEOL(code);
	}
	code->scan = 0;
	if (code->bitnr)
	{
		FaxFlushBuf(code,((7+code->bitnr) >> 3));
	}
}

void FaxInitCoder(CODERSTATE *code)
{
	short i,k;
	MEMClear(sizeof(CODERSTATE),code);
	code->chid=CHANNEL_NONE;
	k = 0x80;
	for(i = 0; i < 8; i++)
	{
		code->mbits[i] = k;
		k = k >> 1;
	}
}

void FaxOpenFile(CODERSTATE *code, Channel chid, short resolution, short packmode, short pix)
{
	if (!code) return;
	if (code)
	{
		code->chid=chid;
		code->res = resolution;
		code->pack = packmode;
		code->pix = pix;
		ClearCoder(code);
	}
}

void FaxFreeCoder(CODERSTATE *code)
{
	if(code->chid!=CHANNEL_NONE)
	{
		FaxCloseFile(code);
	}
}

/* Write QFAX Header. */
/* hd :  File Number */
/* Page: Number of pages in file */

void FaxHeader(CODERSTATE *code, short page, short width, short length, char *tsi)
{
	QFAXHEADER h;
	int pos=0;

	if (code->chid!=CHANNEL_NONE)
	{
		IOPositionSet(code->chid, 0);
		if (page)
		{
			h.title = QFAXTITLE;
			h.version = QFAXVERS;
			h.width = width;
			h.resolution = code->res;
			h.coding = code->pack;
			h.page = length;
			h.pagemax = page;
			if(tsi)
				STRCopy(tsi,h.tsi); 	/* Local TSI, can be blank */
			TIMEGet(&h.stamp);
			h.hint = code->maxlines;	/* Max lines in a page this fax */
		}
		else
		{
			MEMClear(QFAXHEADLEN,&h);	/* place keeper */
		}
		IOPutBytes(code->chid,QFAXHEADLEN,(char *)&h,NULL);
	}
}
