/*
	PROforma, PROGS Font & Raster Manager

	test program for v1.10
	check the speed of - fontcache
					   - mask drawing
					   - positioning strings
	The timing results are written to "ram1_cachetime_txt" (overwrite).

	(c) PROGS, Professional & Graphical Software
	written by Joachim Van der Auwera
	started 5 October 1995
*/

#include "io.h"
#include "sms.h"
#include "job.h"
#include "time.h"
#include "thing.h"

#include "PROforma.h"

const short title1[]={ 'P','R','O','f','o','r','m','a',0x2026,0 };
const short title2[]={ 0xa9,' ','1','9','9','3',0 };
const short title3[]={ 'J','o','a','c','h','i','m',' ',0x26,' ','N','a','t','h','a','n',0 };
const short title4[]={ 0xa3, 0xb1,0xb5,0xbb,0xc0,0xc1,0xc2,0xc3,0xc4,0xc5,0xc5,0xe0,0xe1,0xe2,0xe3,0xe4,0xe5,0xe5,0};

Bbox bbox={short2pt(720), short2pt(540),0,0};

Error TimeDrawing(Gstate,Channel, int count, char *text, char show);

Error init()
{
	Gstate gstate;
	Error err;
	Channel chid;

	err=IOOpen("ram1_cachetime_txt",OPEN_OVER,&chid);
	if (err) return err;

	LinkPROforma();
	err=PFGstateInit(PF_DRIVER_SCREEN,NULL,&bbox,&gstate,NULL);
	if (err) return err;
	err=PFPageShow(gstate);
	if (err) return err;

	err=PFFontLoad(gstate,"Courier");
	if (err) return err;

	/* draw once to fill the fontcache */
	TimeDrawing(gstate,chid,1,"fill the fontcache",TRUE);

	/* the eye also wants something */
	err=PFPageShow(gstate);
	if (err) return err;

	/* test the speed of the fontcache */
	TimeDrawing(gstate,chid,2000,"draw some text",TRUE);

	/* determine size of overhead (no PFShow Operators) */
	TimeDrawing(gstate,chid,2000,"check overhead",FALSE);

	/* use dynamic config to replace bitmap draw by nop routintes */
	/* to determine which part of the time is used by bitmap driver */
	err=THINGCall(PF_THING_NAME,PF_THING_CNFG,1,"D mode4NoDraw_pfd");
	err=PFGstateInit(PF_DRIVER_SCREEN,NULL,&bbox,&gstate,NULL);
	if (err) return err;
	TimeDrawing(gstate,chid,2000,"draw some text without drawing",TRUE);


	return err;
}

Error TimeDrawing(Gstate gstate,Channel chid,int count,char *text,char show)
{
	char str[80],time[30];
	Time before,after;

	STRFormatPut(str,"%s, repeated %d times",text,count);
	IOPutString(chid,str,NULL);
	TIMEGet(&before);
	for (;count>0;count--)
	{
		PFFontScale(gstate, short2pt(80));
		PFMoveTo(gstate, short2pt(20), short2pt(80));
		if (show) PFShow(gstate, title1);
		PFMoveTo(gstate, short2pt(20), short2pt(160));
		if (show) PFShow(gstate, title2);

		PFFontScale(gstate, short2pt(40));
		PFMoveTo(gstate, short2pt(50), short2pt(300));
		if (show) PFShow(gstate, title3);
		PFMoveTo(gstate, short2pt(50), short2pt(340));
		if (show) PFShow(gstate, title4);

		PFFontScale(gstate, short2pt(20));
		PFMoveTo(gstate, short2pt(100), short2pt(400));
		if (show) PFShow(gstate, title1);
		PFMoveTo(gstate, short2pt(100), short2pt(420));
		if (show) PFShow(gstate, title2);
		PFMoveTo(gstate, short2pt(100), short2pt(460));
		if (show) PFShow(gstate, title3);
	}
	TIMEGet(&after);
	STRTime(before,time);
	STRFormatPut(str,"before %s",time);
	IOPutString(chid,str,NULL);
	STRTime(after,time);
	STRFormatPut(str,"after %s",time);
	IOPutString(chid,str,NULL);
	STRFormatPut(str,"time used %d seconds",after-before);
	IOPutString(chid,str,NULL);
	return ERR_OK;
}
